from unittest import TestCase
from mock import Mock, patch, ANY, call

from samcli.local.lambda_service.local_lambda_invoke_service import LocalLambdaInvokeService
from samcli.local.lambdafn.exceptions import FunctionNotFound


class TestLocalLambdaService(TestCase):

    def test_initalize_creates_default_values(self):
        lambda_runner_mock = Mock()
        service = LocalLambdaInvokeService(lambda_runner=lambda_runner_mock, port=3001, host='127.0.0.1')
        self.assertEquals(service.port, 3001)
        self.assertEquals(service.host, '127.0.0.1')
        self.assertEquals(service.lambda_runner, lambda_runner_mock)
        self.assertIsNone(service.stderr)

    def test_initalize_with_values(self):
        lambda_runner_mock = Mock()
        stderr_mock = Mock()
        local_service = LocalLambdaInvokeService(lambda_runner_mock, port=5000, host='129.0.0.0', stderr=stderr_mock)
        self.assertEquals(local_service.port, 5000)
        self.assertEquals(local_service.host, '129.0.0.0')
        self.assertEquals(local_service.stderr, stderr_mock)
        self.assertEquals(local_service.lambda_runner, lambda_runner_mock)

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LocalLambdaInvokeService._construct_error_handling')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.Flask')
    def test_create_service_endpoints(self, flask_mock, error_handling_mock):
        app_mock = Mock()
        flask_mock.return_value = app_mock

        error_handling_mock.return_value = Mock()

        lambda_runner_mock = Mock()
        service = LocalLambdaInvokeService(lambda_runner=lambda_runner_mock, port=3000, host='localhost')

        service.create()

        app_mock.add_url_rule.assert_called_once_with('/2015-03-31/functions/<function_name>/invocations',
                                                      endpoint='/2015-03-31/functions/<function_name>/invocations',
                                                      view_func=service._invoke_request_handler,
                                                      methods=['POST'],
                                                      provide_automatic_options=False)

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LocalLambdaInvokeService.service_response')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaOutputParser')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_invoke_request_handler(self, request_mock, lambda_output_parser_mock, service_response_mock):
        lambda_output_parser_mock.get_lambda_output.return_value = 'hello world', None, False
        service_response_mock.return_value = 'request response'
        request_mock.get_data.return_value = b'{}'

        lambda_runner_mock = Mock()
        service = LocalLambdaInvokeService(lambda_runner=lambda_runner_mock, port=3000, host='localhost')

        response = service._invoke_request_handler(function_name='HelloWorld')

        self.assertEquals(response, 'request response')

        lambda_runner_mock.invoke.assert_called_once_with('HelloWorld', '{}', stdout=ANY, stderr=None)
        service_response_mock.assert_called_once_with('hello world', {'Content-Type': 'application/json'}, 200)

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaErrorResponses')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_invoke_request_handler_on_incorrect_path(self, request_mock, lambda_error_responses_mock):
        request_mock.get_data.return_value = b'{}'
        lambda_runner_mock = Mock()
        lambda_runner_mock.invoke.side_effect = FunctionNotFound

        lambda_error_responses_mock.resource_not_found.return_value = "Couldn't find Lambda"

        service = LocalLambdaInvokeService(lambda_runner=lambda_runner_mock, port=3000, host='localhost')

        response = service._invoke_request_handler(function_name='NotFound')

        self.assertEquals(response, "Couldn't find Lambda")

        lambda_runner_mock.invoke.assert_called_once_with('NotFound', '{}', stdout=ANY, stderr=None)

        lambda_error_responses_mock.resource_not_found.assert_called_once_with('NotFound')

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LocalLambdaInvokeService.service_response')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaOutputParser')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_request_handler_returns_process_stdout_when_making_response(self, request_mock, lambda_output_parser_mock,
                                                                         service_response_mock):
        request_mock.get_data.return_value = b'{}'

        lambda_logs = "logs"
        lambda_response = "response"
        is_customer_error = False
        lambda_output_parser_mock.get_lambda_output.return_value = lambda_response, lambda_logs, is_customer_error

        service_response_mock.return_value = 'request response'

        lambda_runner_mock = Mock()
        stderr_mock = Mock()
        service = LocalLambdaInvokeService(lambda_runner=lambda_runner_mock,
                                           port=3000,
                                           host='localhost',
                                           stderr=stderr_mock)

        result = service._invoke_request_handler(function_name='HelloWorld')

        self.assertEquals(result, 'request response')
        lambda_output_parser_mock.get_lambda_output.assert_called_with(ANY)

        # Make sure the logs are written to stderr
        stderr_mock.write.assert_called_with(lambda_logs)

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaErrorResponses')
    def test_construct_error_handling(self, lambda_error_response_mock):
        service = LocalLambdaInvokeService(lambda_runner=Mock(),
                                           port=3000,
                                           host='localhost',
                                           stderr=Mock())

        flask_app_mock = Mock()
        service._app = flask_app_mock
        service._construct_error_handling()

        flask_app_mock.register_error_handler.assert_has_calls([
            call(500, lambda_error_response_mock.generic_service_exception),
            call(404, lambda_error_response_mock.generic_path_not_found),
            call(405, lambda_error_response_mock.generic_method_not_allowed)])

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LocalLambdaInvokeService.service_response')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaOutputParser')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_invoke_request_handler_with_lambda_that_errors(self,
                                                            request_mock,
                                                            lambda_output_parser_mock,
                                                            service_response_mock):
        lambda_output_parser_mock.get_lambda_output.return_value = 'hello world', None, True
        service_response_mock.return_value = 'request response'
        request_mock.get_data.return_value = b'{}'

        lambda_runner_mock = Mock()
        service = LocalLambdaInvokeService(lambda_runner=lambda_runner_mock, port=3000, host='localhost')

        response = service._invoke_request_handler(function_name='HelloWorld')

        self.assertEquals(response, 'request response')

        lambda_runner_mock.invoke.assert_called_once_with('HelloWorld', '{}', stdout=ANY, stderr=None)
        service_response_mock.assert_called_once_with('hello world',
                                                      {'Content-Type': 'application/json',
                                                       'x-amz-function-error': 'Unhandled'},
                                                      200)

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LocalLambdaInvokeService.service_response')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaOutputParser')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_invoke_request_handler_with_no_data(self, request_mock, lambda_output_parser_mock, service_response_mock):
        lambda_output_parser_mock.get_lambda_output.return_value = 'hello world', None, False
        service_response_mock.return_value = 'request response'
        request_mock.get_data.return_value = None

        lambda_runner_mock = Mock()
        service = LocalLambdaInvokeService(lambda_runner=lambda_runner_mock, port=3000, host='localhost')

        response = service._invoke_request_handler(function_name='HelloWorld')

        self.assertEquals(response, 'request response')

        lambda_runner_mock.invoke.assert_called_once_with('HelloWorld', '{}', stdout=ANY, stderr=None)
        service_response_mock.assert_called_once_with('hello world', {'Content-Type': 'application/json'}, 200)


class TestValidateRequestHandling(TestCase):

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaErrorResponses')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_request_with_non_json_data(self, flask_request, lambda_error_responses_mock):
        flask_request.get_data.return_value = b'notat:asdfasdf'
        flask_request.headers = {}
        flask_request.content_type = 'application/json'
        flask_request.args = {}

        lambda_error_responses_mock.invalid_request_content.return_value = "InvalidRequestContent"

        response = LocalLambdaInvokeService.validate_request()

        self.assertEquals(response, "InvalidRequestContent")

        expected_called_with = "Could not parse request body into json: No JSON object could be decoded"

        lambda_error_responses_mock.invalid_request_content.assert_called_once_with(expected_called_with)

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaErrorResponses')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_request_with_query_strings(self, flask_request, lambda_error_responses_mock):
        flask_request.get_data.return_value = None
        flask_request.headers = {}
        flask_request.content_type = 'application/json'
        flask_request.args = {"key": "value"}

        lambda_error_responses_mock.invalid_request_content.return_value = "InvalidRequestContent"

        response = LocalLambdaInvokeService.validate_request()

        self.assertEquals(response, "InvalidRequestContent")

        lambda_error_responses_mock.invalid_request_content.assert_called_once_with(
            "Query Parameters are not supported")

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaErrorResponses')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_request_log_type_not_None(self, flask_request, lambda_error_responses_mock):
        flask_request.get_data.return_value = None
        flask_request.headers = {'X-Amz-Log-Type': 'Tail'}
        flask_request.content_type = 'application/json'
        flask_request.args = {}

        lambda_error_responses_mock.not_implemented_locally.return_value = "NotImplementedLocally"

        response = LocalLambdaInvokeService.validate_request()

        self.assertEquals(response, "NotImplementedLocally")

        lambda_error_responses_mock.not_implemented_locally.assert_called_once_with(
            "log-type: Tail is not supported. None is only supported.")

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.LambdaErrorResponses')
    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_request_invocation_type_not_ResponseRequest(self, flask_request, lambda_error_responses_mock):
        flask_request.get_data.return_value = None
        flask_request.headers = {'X-Amz-Invocation-Type': 'DryRun'}
        flask_request.content_type = 'application/json'
        flask_request.args = {}

        lambda_error_responses_mock.not_implemented_locally.return_value = "NotImplementedLocally"

        response = LocalLambdaInvokeService.validate_request()

        self.assertEquals(response, "NotImplementedLocally")

        lambda_error_responses_mock.not_implemented_locally.assert_called_once_with(
            "invocation-type: DryRun is not supported. RequestResponse is only supported.")

    @patch('samcli.local.lambda_service.local_lambda_invoke_service.request')
    def test_request_with_no_data(self, flask_request):
        flask_request.get_data.return_value = None
        flask_request.headers = {}
        flask_request.content_type = 'application/json'
        flask_request.args = {}

        response = LocalLambdaInvokeService.validate_request()

        self.assertIsNone(response)
